import ot
import numpy as np
import scipy
import networkx


def OT_matching(sim_matrix, entropy_reg=0.01):
    weights = sim_matrix

    # Negative weights are interpreted as costs, scale them to be in [0, 1]
    C = -weights + 1
    C -= C.min()
    C += 1
    C = C / C.max()
    del weights

    # Uniform distributions on cell profiles
    mod1_distr = np.ones(sim_matrix.shape[0]) / sim_matrix.shape[0]
    mod2_distr = np.ones(sim_matrix.shape[1]) / sim_matrix.shape[1]

    if entropy_reg > 0:
        # Entropy-regularized OT
        bipartite_matching_adjacency = ot.sinkhorn(
            mod1_distr, mod2_distr, reg=0.01, M=C, numItermax=5000, verbose=True
        )
    else:
        # Exact OT
        bipartite_matching_adjacency = ot.emd(
            mod1_distr, mod2_distr, M=C, numItermax=10000000
        )

    return bipartite_matching_adjacency


def MWB_matching(raw_logits, threshold_quantile=0.99):
    while threshold_quantile >= 0:
        try:
            weights = raw_logits.copy()
            # Discard weights smaller than the threshold quantile
            quantile_row = np.quantile(
                weights, threshold_quantile, axis=0, keepdims=True
            )
            quantile_col = np.quantile(
                weights, threshold_quantile, axis=1, keepdims=True
            )
            mask_ = weights < quantile_row
            mask_ = np.logical_and(mask_, (weights < quantile_col), out=mask_)
            weights[mask_] = 0
            weights_sparse = scipy.sparse.csr_matrix(-weights)
            del weights
            graph = networkx.algorithms.bipartite.matrix.from_biadjacency_matrix(
                weights_sparse
            )
            u = [n for n in graph.nodes if graph.nodes[n]["bipartite"] == 0]
            matches = (
                networkx.algorithms.bipartite.matching.minimum_weight_full_matching(
                    graph, top_nodes=u
                )
            )
            best_matches = np.array([matches[x] - len(u) for x in u])
            bipartite_matching_adjacency = np.zeros(raw_logits.shape)
            bipartite_matching_adjacency[
                np.arange(raw_logits.shape[0]), best_matches
            ] = 1
            return bipartite_matching_adjacency
        except networkx.exception.NetworkXException:
            threshold_quantile -= 0.01
            print(
                "Discarding too many edges, problem infeasible, lowering threshold to {}".format(
                    threshold_quantile
                )
            )
    raise RuntimeError("No feasible solution found")



import torch
from torch.utils.data import Dataset, DataLoader


class ModalityMatchingDataset(Dataset):
    def __init__(self, df_modality1, df_modality2):
        super().__init__()

        self.df_modality1 = df_modality1.values
        self.df_modality2 = df_modality2.values

    def __len__(self):
        return self.df_modality1.shape[0]

    def __getitem__(self, index: int):
        x_modality_1 = self.df_modality1[index]
        x_modality_2 = self.df_modality2[index]
        return {"features_first": x_modality_1, "features_second": x_modality_2}


def get_dataloaders(
    mod1_train,
    mod2_train,
    sol_train,
    mod1_test,
    mod2_test,
    sol_test,
    NUM_WORKERS,
    BATCH_SIZE,
):
    mod2_train = mod2_train.iloc[sol_train.values.argmax(1)]
    mod2_test = mod2_test.iloc[sol_test.values.argmax(1)]

    dataset_train = ModalityMatchingDataset(mod1_train, mod2_train)
    data_train = DataLoader(
        dataset_train, BATCH_SIZE, shuffle=True, num_workers=NUM_WORKERS
    )

    dataset_test = ModalityMatchingDataset(mod1_test, mod2_test)
    data_test = DataLoader(
        dataset_test, BATCH_SIZE, shuffle=False, num_workers=NUM_WORKERS
    )

    return data_train, data_test


class AugDataset(Dataset):
    """
    Modality matching dataset class for training with affine transformation data augmentations. Given 4 affine
    transformations for each modality, it applies a random transformation which is an interpolation of the 4
    transformations. Can be used to augment a distribution of cell profiles from a single batch by applying affine
    transformations that approximate the optimal transport map between the non-augmented batch and other batches.
    """

    def __init__(
        self,
        df_modality1,
        df_modality2,
        transf_matrices_mod1,
        transf_vectors_mod1,
        transf_matrices_mod2,
        transf_vectors_mod2,
    ):
        """
        Args:
            df_modality1: pandas dataframe of mod1 data
            df_modality2: pandas dataframe of mod2 data
            transf_matrices_mod1: list of 4 linear transformation matrices for augmenting mod1
            transf_vectors_mod1: list of 4 translation vectors for augmenting mod1
            transf_matrices_mod2: list of 4 linear transformation matrices for augmenting mod2
            transf_vectors_mod2: list of 4 translation vectors for augmenting mod2
        """
        super().__init__()

        self.df_modality1 = df_modality1.values
        self.df_modality2 = df_modality2.values

        # save the affine transformations
        self.transf_matrices_mod1 = transf_matrices_mod1
        self.transf_vectors_mod1 = transf_vectors_mod1
        self.transf_matrices_mod2 = transf_matrices_mod2
        self.transf_vectors_mod2 = transf_vectors_mod2

        # define the tetrahedron
        n_dim = 3
        r1 = [1, 1, 1]
        r2 = [1, -1, -1]
        r3 = [-1, 1, -1]
        r4 = [-1, -1, 1]
        rfirst = [r1, r2, r3]
        # transformation matrix to get the barycentric coordinates
        T = torch.Tensor(
            [
                [rfirst[col][row] - r4[row] for col in range(n_dim)]
                for row in range(n_dim)
            ]
        )
        Tinv = torch.linalg.inv(T)
        # class attributes used by self._sample_mixing_coeff()
        self.r4 = torch.Tensor(r4)
        self.Tinv = Tinv

    def _sample_mixing_coeff(self):
        """
        Returns 4 mixing coefficients, which are the barycentric coordinates of a random point sampled through
        rejection sampling from a uniform distribution over the volume of a regular 3D tetrahedron. Coordinate
        conversion formulas from https://en.wikipedia.org/wiki/Barycentric_coordinate_system
        #Barycentric_coordinates_on_tetrahedra
        """
        # this loop has probability of exiting 1/3 (= tetrahedron volume / cube volume) at each iteration
        # therefore, the expected running time is 3 iterations
        while True:
            # sample a point from the uniform distribution on [-1, 1]^3
            pt = 2 * torch.rand(3) - 1
            # reject and resample the point if it's outside the tetrahedron
            if not (
                pt[0] < pt[1] + pt[2] - 1
                or pt[0] > -pt[1] + pt[2] + 1
                or pt[0] > pt[1] - pt[2] + 1
                or pt[0] < -pt[1] - pt[2] - 1
            ):
                break

        # transform from cartesian to barycentric coordinates
        c1, c2, c3 = (self.Tinv @ (pt - self.r4)).tolist()
        c4 = 1 - c1 - c2 - c3

        return c1, c2, c3, c4

    def __len__(self):
        return self.df_modality1.shape[0]

    def __getitem__(self, index: int):
        x_mod2 = self.df_modality1[index]
        x_mod1 = self.df_modality2[index]

        c1, c2, c3, c4 = self._sample_mixing_coeff()
        # sample a random augmentation strength uniformly between 0 and 1
        aug_strength = torch.rand(size=(1,))
        # use the same coefficients for the two modalities so the matching still makes sense
        x_mod1 = (
            aug_strength
            * (
                c1
                * (self.transf_matrices_mod1[0] @ x_mod1 + self.transf_vectors_mod1[0])
                + c2
                * (self.transf_matrices_mod1[1] @ x_mod1 + self.transf_vectors_mod1[1])
                + c3
                * (self.transf_matrices_mod1[2] @ x_mod1 + self.transf_vectors_mod1[2])
                + c4
                * (self.transf_matrices_mod1[3] @ x_mod1 + self.transf_vectors_mod1[3])
            )
            + (1 - aug_strength) * x_mod1
        )
        x_mod2 = (
            aug_strength
            * (
                c1
                * (self.transf_matrices_mod2[0] @ x_mod2 + self.transf_vectors_mod2[0])
                + c2
                * (self.transf_matrices_mod2[1] @ x_mod2 + self.transf_vectors_mod2[1])
                + c3
                * (self.transf_matrices_mod2[2] @ x_mod2 + self.transf_vectors_mod2[2])
                + c4
                * (self.transf_matrices_mod2[3] @ x_mod2 + self.transf_vectors_mod2[3])
            )
            + (1 - aug_strength) * x_mod2
        )
        return {"features_first": x_mod2.squeeze(), "features_second": x_mod1.squeeze()}


class SingleAugDataset(Dataset):
    """
    Similar to AugDataset class, but only applies a single affine transformation to each modality.
    """

    def __init__(
        self,
        df_modality1,
        df_modality2,
        transf_matrix_mod1,
        transf_vector_mod1,
        transf_matrix_mod2,
        transf_vector_mod2,
    ):
        """
        Args:
            df_modality1: pandas dataframe of mod1 data
            df_modality2: pandas dataframe of mod2 data
            transf_matrices_mod1: linear transformation matrix for augmenting mod1
            transf_vectors_mod1: translation vector for augmenting mod1
            transf_matrices_mod2: linear transformation matrix for augmenting mod2
            transf_vectors_mod2: translation vector for augmenting mod2
        """
        super().__init__()

        self.df_modality1 = df_modality1.values
        self.df_modality2 = df_modality2.values

        # save the affine transformations
        self.transf_matrix_mod1 = transf_matrix_mod1
        self.transf_vector_mod1 = transf_vector_mod1
        self.transf_matrix_mod2 = transf_matrix_mod2
        self.transf_vector_mod2 = transf_vector_mod2

    def __len__(self):
        return self.df_modality1.shape[0]

    def __getitem__(self, index: int):
        x_mod2 = self.df_modality1[index]
        x_mod1 = self.df_modality2[index]

        # sample a random augmentation strength uniformly between 0 and 1
        aug_strength = torch.rand(size=(1,))
        # use the same coefficients for the two modalities so the matching still makes sense
        x_mod1 = (
            aug_strength * (self.transf_matrix_mod1 @ x_mod1 + self.transf_vector_mod1)
            + (1 - aug_strength) * x_mod1
        )
        x_mod2 = (
            aug_strength * (self.transf_matrix_mod2 @ x_mod2 + self.transf_vector_mod2)
            + (1 - aug_strength) * x_mod2
        )

        return {"features_first": x_mod2.squeeze(), "features_second": x_mod1.squeeze()}



# Default args and hyperparameters
defaults_common = dict(
    DATASETS_PATH="datasets",
    PRETRAIN_PATH="pretrain",
    OUT_NAME="",
    SCORES_PATH="scores",
    VALID_FOLD=0,  # validation on the first batch (s1d1), train on other batches
    HYPERPARAMS=True,
    OT_MATCHING=True,
    BATCH_LABEL_MATCHING=True,
    OT_ENTROPY=0.01,
    TRANSDUCTIVE=True,
    HARMONY=True,
    CUSTOM_DATASET_PATH="",
    SEED=0,
    SAVE_EMBEDDINGS=False,
)

defaults_GEX2ATAC = dict(
    LR=0.0006,
    WEIGHT_DECAY=0.000125,
    EMBEDDING_DIM=128,
    DROPOUT_RATES_ATAC=0.67,
    DROPOUT_RATES_GEX0=0.34,
    DROPOUT_RATES_GEX1=0.47,
    LAYERS_DIM_ATAC=2048,
    LAYERS_DIM_GEX0=2048,
    LAYERS_DIM_GEX1=1024,
    LOG_T=2.74,
    N_LSI_COMPONENTS_ATAC=256,
    N_LSI_COMPONENTS_GEX=192,
    N_EPOCHS=7000,
    BATCH_SIZE=16384,
    SFA_NOISE=0.0,
)

defaults_GEX2ADT = dict(
    LR=0.000175,
    WEIGHT_DECAY=0.0002,
    EMBEDDING_DIM=256,
    DROPOUT_RATES_ADT0=0.4,
    DROPOUT_RATES_ADT1=0.2,
    DROPOUT_RATES_GEX0=0.3,
    DROPOUT_RATES_GEX1=0.05,
    LAYERS_DIM_ADT0=4096,
    LAYERS_DIM_ADT1=2048,
    LAYERS_DIM_GEX0=256,
    LAYERS_DIM_GEX1=2048,
    LOG_T=4.0,
    N_LSI_COMPONENTS_GEX=192,
    N_LSI_COMPONENTS_ADT=134,  # ADT data is not processed with LSI, just used for the encoder input dimension
    N_EPOCHS=7000,
    BATCH_SIZE=16384,
    SFA_NOISE=0.0,
)

# Team Novel baseline hyperparameters
baseline_GEX2ATAC = dict(
    LR=0.000585,
    WEIGHT_DECAY=0.0,
    EMBEDDING_DIM=256,
    DROPOUT_RATES_ATAC=0.661,
    DROPOUT_RATES_GEX0=0.541,
    DROPOUT_RATES_GEX1=0.396,
    LAYERS_DIM_ATAC=2048,
    LAYERS_DIM_GEX0=1024,
    LAYERS_DIM_GEX1=1024,
    LOG_T=3.065016,
    N_LSI_COMPONENTS_ATAC=512,
    N_LSI_COMPONENTS_GEX=64,
    N_EPOCHS=7000,
    BATCH_SIZE=16384,
    SFA_NOISE=0.0,
)

baseline_GEX2ADT = dict(
    LR=7.79984e-05,
    WEIGHT_DECAY=0.0,
    EMBEDDING_DIM=64,
    DROPOUT_RATES_ADT0=0.0221735,
    DROPOUT_RATES_ADT1=0.296919,
    DROPOUT_RATES_GEX0=0.0107121,
    DROPOUT_RATES_GEX1=0.254689,
    LAYERS_DIM_ADT0=512,
    LAYERS_DIM_ADT1=2048,
    LAYERS_DIM_GEX0=1024,
    LAYERS_DIM_GEX1=512,
    LOG_T=3.463735,
    N_LSI_COMPONENTS_GEX=128,
    N_LSI_COMPONENTS_ADT=134,  # ADT data is not processed with LSI, just used for the encoder input dimension
    N_EPOCHS=7000,
    BATCH_SIZE=2048,
    SFA_NOISE=0.0,
)



import os
import anndata as ad
import torch
import numpy
import argparse


def evaluate(prediction_test, sol_test, scores_path="scores/scores.txt"):
    os.makedirs(os.path.dirname(scores_path), exist_ok=True)
    with open(scores_path, "a+") as f:
        if type(prediction_test.X) != numpy.ndarray:
            X = prediction_test.X.toarray()
        else:
            X = prediction_test.X
        X = torch.tensor(X)

        Xsol = torch.tensor(sol_test.X.toarray())
        Xsol.argmax(1)
        # Order the columns of the prediction matrix so that the perfect prediction is the identity matrix
        X = X[:, Xsol.argmax(1)]

        labels = torch.arange(X.shape[0])
        forward_accuracy = (torch.argmax(X, dim=1) == labels).float().mean().item()
        backward_accuracy = (torch.argmax(X, dim=0) == labels).float().mean().item()
        avg_accuracy = 0.5 * (forward_accuracy + backward_accuracy)
        print("top1 forward acc:", forward_accuracy)
        print("top1 forward acc:", forward_accuracy, file=f)
        print("top1 backward acc:", backward_accuracy)
        print("top1 backward acc:", backward_accuracy, file=f)
        print("top1 avg acc:", avg_accuracy)
        print("top1 avg acc:", avg_accuracy, file=f)

        _, top_indexes_forward = X.topk(5, dim=1)
        _, top_indexes_backward = X.topk(5, dim=0)
        l_forward = labels.expand(5, X.shape[0]).T
        l_backward = l_forward.T
        top5_forward_accuracy = (
            torch.any(top_indexes_forward == l_forward, 1).float().mean().item()
        )
        top5_backward_accuracy = (
            torch.any(top_indexes_backward == l_backward, 0).float().mean().item()
        )
        top5_avg_accuracy = 0.5 * (top5_forward_accuracy + top5_backward_accuracy)

        print("top5 forward acc:", top5_forward_accuracy)
        print("top5 forward acc:", top5_forward_accuracy, file=f)
        print("top5 backward acc:", top5_backward_accuracy)
        print("top5 backward acc:", top5_backward_accuracy, file=f)
        print("top5 avg acc:", top5_avg_accuracy)
        print("top5 avg acc:", top5_avg_accuracy, file=f)

        logits_row_sums = X.clip(min=0).sum(dim=1)
        top1_competition_metric = (
            X.clip(min=0).diagonal().div(logits_row_sums).mean().item()
        )
        print("top1 competition metric:", top1_competition_metric)
        print("top1 competition metric:", top1_competition_metric, file=f)

        # For soft predictions, the competition score can be made equal to the forward accuracy (or backward accuracy) by
        # putting 1 at the max of each row (or each column) and 0 elsewhere
        mx = torch.max(X, dim=1, keepdim=True).values
        hard_X = (mx == X).float()
        logits_row_sums = hard_X.clip(min=0).sum(dim=1)
        top1_competition_metric = (
            hard_X.clip(min=0).diagonal().div(logits_row_sums).mean().item()
        )
        print("top1 competition metric for soft predictions:", top1_competition_metric)
        print(
            "top1 competition metric for soft predictions:",
            top1_competition_metric,
            file=f,
        )

        # FOSCTTM
        foscttm = (X > torch.diag(X)).float().mean().item()
        foscttm_x = (X >= torch.diag(X)).float().mean(axis=1).mean().item()
        foscttm_y = (X >= torch.diag(X)).float().mean(axis=0).mean().item()
        print("foscttm:", foscttm, "foscttm_x:", foscttm_x, "foscttm_y:", foscttm_y)
        print(
            "foscttm:",
            foscttm,
            "foscttm_x:",
            foscttm_x,
            "foscttm_y:",
            foscttm_y,
            file=f,
        )


if __name__ == "__main__":
    os.environ["CUDA_LAUNCH_BLOCKING"] = "1"
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="TASK")
    parser_GEX2ADT = subparsers.add_parser("GEX2ADT", help="GEX2ADT model")
    parser_GEX2ATAC = subparsers.add_parser("GEX2ATAC", help="GEX2ATAC model")
    parser.add_argument("--prediction_path", "-p", type=str, default="GEX2ADT.h5ad")
    parser.add_argument("--dataset_path", "-d", type=str, default="datasets")
    args, unknown_args = parser.parse_known_args()

    if args.TASK == "GEX2ADT":
        test_path = os.path.join(
            args.dataset_path,
            "openproblems_bmmc_cite_phase2_rna/openproblems_bmmc_cite_phase2_rna"
            ".censor_dataset.output_",
        )
    elif args.TASK == "GEX2ATAC":
        test_path = os.path.join(
            args.dataset_path,
            "openproblems_bmmc_multiome_phase2_rna"
            "/openproblems_bmmc_multiome_phase2_rna.censor_dataset.output_",
        )
    else:
        raise ValueError("Unknown task: " + args.TASK)

    par = {
        "input_test_prediction": args.prediction_path,
        "input_test_sol": f"{test_path}test_sol.h5ad",
    }

    prediction_test = ad.read_h5ad(par["input_test_prediction"])
    sol_test = ad.read_h5ad(par["input_test_sol"])

    scores_path = os.path.join("scores", args.TASK + ".txt")
    evaluate(prediction_test, sol_test, scores_path)



import argparse
import os
import pickle

import anndata as ad
import pandas as pd
import torch
from catalyst import dl
from catalyst.utils import set_global_seed
from sklearn.model_selection import LeaveOneGroupOut
from distutils.util import strtobool

# sys.path.append(".")
from matchclot.embedding.catalyst_tools import scRNARunner, CustomMetric
from matchclot.embedding.models import Modality_CLIP, Encoder
from matchclot.preprocessing.preprocess import lsiTransformer, harmony
from matchclot.utils.dataloaders import ModalityMatchingDataset
from matchclot.utils.hyperparameters import (
    defaults_common,
    defaults_GEX2ADT,
    defaults_GEX2ATAC,
    baseline_GEX2ATAC,
    baseline_GEX2ADT,
)

if __name__ == "__main__":

    os.environ["CUDA_LAUNCH_BLOCKING"] = "1"
    # setting device on GPU if available, else CPU
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print("Using device:", device)

    # Define argument parsers
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="TASK")

    # Common args
    for key, value in defaults_common.items():
        parser.add_argument(
            "--" + key,
            default=value,
            type=(lambda x: bool(strtobool(x))) if type(value) == bool else type(value),
        )

    # GEX2ADT args
    parser_GEX2ADT = subparsers.add_parser("GEX2ADT", help="train GEX2ADT model")
    for key, value in defaults_GEX2ADT.items():
        parser_GEX2ADT.add_argument("--" + key, default=value, type=type(value))

    # GEX2ATAC args
    parser_GEX2ATAC = subparsers.add_parser("GEX2ATAC", help="train GEX2ATAC model")
    for key, value in defaults_GEX2ATAC.items():
        parser_GEX2ATAC.add_argument("--" + key, default=value, type=type(value))

    # Parse args
    args, unknown_args = parser.parse_known_args()

    # Set global random seed
    set_global_seed(args.SEED)

    # Date in format YYMMDDHHMMSS
    date = "".join(
        [
            c if c.isnumeric() else ""
            for c in str(pd.Timestamp("today").to_pydatetime())
        ][2:19]
    )

    # Define file paths
    if args.TASK == "GEX2ADT":
        dataset_path = os.path.join(
            args.DATASETS_PATH,
            "openproblems_bmmc_cite_phase2_rna/openproblems_bmmc_cite_phase2_rna.censor_dataset.output_",
        )
        pretrain_path = os.path.join(
            args.PRETRAIN_PATH, "GEX2ADT"
        )  # Path for saving the trained model
        is_multiome = False
    elif args.TASK == "GEX2ATAC":
        dataset_path = os.path.join(
            args.DATASETS_PATH,
            "openproblems_bmmc_multiome_phase2_rna/openproblems_bmmc_multiome_phase2_rna.censor_dataset.output_",
        )
        pretrain_path = os.path.join(args.PRETRAIN_PATH, "GEX2ATAC")
        is_multiome = True
    else:
        raise ValueError("Unknown task: " + args.TASK)

    par = {
        "input_train_mod1": f"{dataset_path}train_mod1.h5ad",
        "input_train_mod2": f"{dataset_path}train_mod2.h5ad",
        "input_train_sol": f"{dataset_path}train_sol.h5ad",
        "input_test_mod1": f"{dataset_path}test_mod1.h5ad",
        "input_test_mod2": f"{dataset_path}test_mod2.h5ad",
        "output_pretrain": pretrain_path,
        "input_pretrain": pretrain_path,
    }
    os.makedirs(par["output_pretrain"], exist_ok=True)

    # Overwrite configurations for ablation study
    if args.HYPERPARAMS is False:
        if is_multiome:
            for hyperparam, baseline_value in baseline_GEX2ATAC.items():
                setattr(args, hyperparam, baseline_value)
        else:
            for hyperparam, baseline_value in baseline_GEX2ADT.items():
                setattr(args, hyperparam, baseline_value)
    print("args:", args, "unknown_args:", unknown_args)

    # Load train data
    print("loading train data")
    input_train_mod1 = ad.read_h5ad(par["input_train_mod1"])
    print("input_train_mod1.shape", input_train_mod1.shape)
    input_train_mod2 = ad.read_h5ad(par["input_train_mod2"])
    print("input_train_mod2.shape", input_train_mod2.shape)
    sol_train = ad.read_h5ad(par["input_train_sol"])  # ground truth matching

    mod1 = input_train_mod1.var["feature_types"][0]
    mod2 = input_train_mod2.var["feature_types"][0]
    assert mod1 == "GEX"  # mod1 is always GEX, mod2 is either ADT or ATAC

    # Apply the same ordering of mod2 profiles as mod1
    input_train_mod2 = input_train_mod2[sol_train.to_df().values.argmax(1)]

    # Load private test data, used for transductive LSI + Harmony preprocessing
    print("loading private test data")
    input_test_mod1 = ad.read_h5ad(par["input_test_mod1"])
    print("input_test_mod1.shape", input_test_mod1.shape)
    input_test_mod2 = ad.read_h5ad(par["input_test_mod2"])
    print("input_test_mod2.shape", input_test_mod2.shape)

    # Define train and validation split
    fold_number = args.VALID_FOLD
    print("fold_number:", fold_number)
    trial_dump_folder = os.path.join(par["output_pretrain"], str(fold_number))
    logo = LeaveOneGroupOut()
    groups = sol_train.obs.batch
    print("GROUPS:", groups)
    logo.get_n_splits(input_train_mod2, groups=groups)
    all_splits = list(logo.split(input_train_mod2, groups=groups))
    train_indexes, test_indexes = all_splits[fold_number]
    print("len train:", len(train_indexes), "len test:", len(test_indexes))

    # Load or fit LSI preprocessing
    path = par["output_pretrain"]

    if (
        os.path.exists(path + "/lsi_GEX_transformer.pickle")
        and args.TRANSDUCTIVE
        and not is_multiome
    ):
        # Avoid re-computing LSI transformation when using cross-validation and transductive LSI
        print("loading lsi transformer from", path)
        # LSI is applied only on GEX and ATAC, not on ADT
        with open(path + "/lsi_GEX_transformer.pickle", "rb") as f:
            lsi_transformer_gex = pickle.load(f)
    elif (
        os.path.exists(path + "/lsi_GEX_transformer.pickle")
        and os.path.exists(path + "/lsi_ATAC_transformer.pickle")
        and args.TRANSDUCTIVE
        and is_multiome
    ):
        with open(path + "/lsi_GEX_transformer.pickle", "rb") as f:
            lsi_transformer_gex = pickle.load(f)
        with open(path + "/lsi_ATAC_transformer.pickle", "rb") as f:
            lsi_transformer_atac = pickle.load(f)
    else:
        print("No lsi transformer found in", path, "creating new one")
        os.makedirs(path, exist_ok=True)

        # Fit GEX LSI
        lsi_transformer_gex = lsiTransformer(
            n_components=args.N_LSI_COMPONENTS_GEX, drop_first=True
        )
        if args.TRANSDUCTIVE:
            print("concatenating gex train and test")
            concatenated_gex = ad.concat(
                [input_train_mod1, input_test_mod1], join="outer"
            )
            print("done, concatenated_gex.shape", concatenated_gex.shape)
            lsi_transformer_gex.fit(concatenated_gex)
            # Save LSI transformation
            with open(path + "/lsi_GEX_transformer.pickle", "wb") as f:
                pickle.dump(lsi_transformer_gex, f)
            print("saved lsi pickle in ", path + "/lsi_GEX_transformer.pickle")
        else:
            lsi_transformer_gex.fit(input_train_mod1)
            with open(path + "/lsi_GEX_transformer.pickle", "wb") as f:
                pickle.dump(lsi_transformer_gex, f)
            print(
                "saved lsi pickle in ",
                trial_dump_folder + "/lsi_GEX_transformer.pickle",
            )

        # LSI is applied only on GEX and ATAC, not on ADT
        if is_multiome:
            # Fit ATAC LSI
            lsi_transformer_atac = lsiTransformer(
                n_components=args.N_LSI_COMPONENTS_ATAC, drop_first=True
            )
            if args.TRANSDUCTIVE:
                print("concatenating atac train and test")
                concatenated_atac = ad.concat(
                    [input_train_mod2, input_test_mod2], join="outer"
                )
                print("done, concatenated_atac.shape", concatenated_atac.shape)
                lsi_transformer_atac.fit(concatenated_atac)
            else:
                lsi_transformer_atac.fit(input_train_mod2)

            # Save LSI transformation
            with open(path + "/lsi_ATAC_transformer.pickle", "wb") as f:
                pickle.dump(lsi_transformer_atac, f)
            print("saved lsi pickle in ", path + "/lsi_ATAC_transformer.pickle")

    # Apply LSI preprocessing
    gex_train = lsi_transformer_gex.transform(input_train_mod1[train_indexes])
    gex_test = lsi_transformer_gex.transform(input_train_mod1[test_indexes])
    gex_private = lsi_transformer_gex.transform(input_test_mod1)
    if is_multiome:
        mod2_train = lsi_transformer_atac.transform(input_train_mod2[train_indexes])
        mod2_test = lsi_transformer_atac.transform(input_train_mod2[test_indexes])
        mod2_private = lsi_transformer_atac.transform(input_test_mod2)
    else:
        mod2_train = input_train_mod2[train_indexes].to_df()
        mod2_test = input_train_mod2[test_indexes].to_df()
        mod2_private = input_test_mod2.to_df()

    # Apply Harmony batch effect correction
    if args.HARMONY:
        gex_train["batch"] = input_train_mod1.obs.batch[train_indexes]
        gex_test["batch"] = input_train_mod1.obs.batch[test_indexes]
        gex_private["batch"] = input_test_mod1.obs.batch
        mod2_train["batch"] = input_train_mod2.obs.batch[train_indexes]
        mod2_test["batch"] = input_train_mod2.obs.batch[test_indexes]
        mod2_private["batch"] = input_test_mod2.obs.batch

        if args.TRANSDUCTIVE:
            # Transductive setting
            gex_train, gex_test, gex_private = harmony(
                [gex_train, gex_test, gex_private]
            )
            mod2_train, mod2_test, mod2_private = harmony(
                [mod2_train, mod2_test, mod2_private]
            )
        else:
            (gex_train,) = harmony([gex_train])
            (gex_test,) = harmony([gex_test])
            (mod2_train,) = harmony([mod2_train])
            (mod2_test,) = harmony([mod2_test])

    # Load torch dataloaders
    dataset_train = ModalityMatchingDataset(
        pd.DataFrame(mod2_train), pd.DataFrame(gex_train)
    )
    dataset_test = ModalityMatchingDataset(
        pd.DataFrame(mod2_test), pd.DataFrame(gex_test)
    )
    dataloader_train = torch.utils.data.DataLoader(
        dataset_train, args.BATCH_SIZE, shuffle=True, num_workers=4
    )
    dataloader_test = torch.utils.data.DataLoader(
        dataset_test, 256, shuffle=False, num_workers=4
    )
    print("loaded dataloaders")

    # Define modality encoders and trainer
    if is_multiome:
        model = Modality_CLIP(
            Encoder=Encoder,
            layers_dims=(
                [args.LAYERS_DIM_ATAC],
                [args.LAYERS_DIM_GEX0, args.LAYERS_DIM_GEX1],
            ),
            dropout_rates=(
                [args.DROPOUT_RATES_ATAC],
                [args.DROPOUT_RATES_GEX0, args.DROPOUT_RATES_GEX1],
            ),
            dim_mod1=args.N_LSI_COMPONENTS_ATAC,
            dim_mod2=args.N_LSI_COMPONENTS_GEX,
            output_dim=args.EMBEDDING_DIM,
            T=args.LOG_T,
            noise_amount=args.SFA_NOISE,
        )
    else:
        model = Modality_CLIP(
            Encoder=Encoder,
            layers_dims=(
                [args.LAYERS_DIM_ADT0, args.LAYERS_DIM_ADT1],
                [args.LAYERS_DIM_GEX0, args.LAYERS_DIM_GEX1],
            ),
            dropout_rates=(
                [args.DROPOUT_RATES_ADT0, args.DROPOUT_RATES_ADT1],
                [args.DROPOUT_RATES_GEX0, args.DROPOUT_RATES_GEX1],
            ),
            dim_mod1=args.N_LSI_COMPONENTS_ADT,
            dim_mod2=args.N_LSI_COMPONENTS_GEX,
            output_dim=args.EMBEDDING_DIM,
            T=args.LOG_T,
            noise_amount=args.SFA_NOISE,
        )

    optimizer = torch.optim.Adam(
        model.parameters(), args.LR, weight_decay=args.WEIGHT_DECAY
    )
    loaders = {
        "train": dataloader_train,
        "valid": dataloader_test,
    }
    runner = scRNARunner()

    # Train model
    runner.train(
        model=model,
        optimizer=optimizer,
        loaders=loaders,
        num_epochs=args.N_EPOCHS,
        callbacks=[
            dl.OptimizerCallback(metric_key="loss"),
            dl.CheckpointCallback(
                logdir=trial_dump_folder,
                loader_key="valid",
                metric_key="avg_acc",
                minimize=False,
                topk=1,
            ),
            dl.EarlyStoppingCallback(
                patience=150,
                loader_key="valid",
                metric_key="avg_acc",
                minimize=False,
                min_delta=1e-5,
            ),
            dl.ControlFlowCallbackWrapper(
                base_callback=dl.LoaderMetricCallback(
                    metric=CustomMetric(),
                    input_key=["embeddings_first", "embeddings_second", "temperature"],
                    target_key=["embeddings_second"],
                ),
                ignore_loaders="train",  # Compute metrics only for validation, takes a long time on the training set
            ),
        ],
        verbose=True,
    )



from typing import Optional

import anndata
import numpy as np
import pandas as pd
import scipy.sparse
import sklearn.decomposition
import sklearn.feature_extraction.text
import sklearn.neighbors
import sklearn.preprocessing
import sklearn.utils.extmath
from harmony import harmonize


def tfidf(X):
    """
    TF-IDF normalization (following the Seurat v3 approach)
    Parameters
    ----------
    X
        Input matrix
    Returns
    -------
    X_tfidf
        TF-IDF normalized matrix
    """
    idf = X.shape[0] / X.sum(axis=0)
    if scipy.sparse.issparse(X):
        tf = X.multiply(1 / X.sum(axis=1))
        return tf.multiply(idf)
    else:
        tf = X / X.sum(axis=1, keepdims=True)
        return tf * idf


class tfidfTransformer:
    def __init__(self):
        self.idf = None
        self.fitted = False

    def fit(self, X):
        self.idf = X.shape[0] / X.sum(axis=0)
        self.fitted = True

    def transform(self, X):
        if not self.fitted:
            raise RuntimeError("Transformer was not fitted on any data")
        if scipy.sparse.issparse(X):
            tf = X.multiply(1 / X.sum(axis=1))
            return tf.multiply(self.idf)
        else:
            tf = X / X.sum(axis=1, keepdims=True)
            return tf * self.idf

    def fit_transform(self, X):
        self.fit(X)
        return self.transform(X)


class lsiTransformer:
    def __init__(
        self, n_components: int = 20, drop_first=True, use_highly_variable=None
    ):

        self.drop_first = drop_first
        self.n_components = n_components + drop_first
        self.use_highly_variable = use_highly_variable
        self.tfidfTransformer = tfidfTransformer()
        self.normalizer = sklearn.preprocessing.Normalizer(norm="l1")
        self.pcaTransformer = sklearn.decomposition.TruncatedSVD(
            n_components=self.n_components, random_state=777
        )
        self.fitted = None

    def fit(self, adata: anndata.AnnData):
        if self.use_highly_variable is None:
            self.use_highly_variable = "highly_variable" in adata.var
        adata_use = (
            adata[:, adata.var["highly_variable"]]
            if self.use_highly_variable
            else adata
        )
        X = self.tfidfTransformer.fit_transform(adata_use.X)
        X_norm = self.normalizer.fit_transform(X)
        X_norm = np.log1p(X_norm * 1e4)
        self.pcaTransformer.fit(X_norm)
        self.fitted = True

    def transform(self, adata):
        if not self.fitted:
            raise RuntimeError("Transformer was not fitted on any data")
        adata_use = (
            adata[:, adata.var["highly_variable"]]
            if self.use_highly_variable
            else adata
        )
        X = self.tfidfTransformer.transform(adata_use.X)
        X_norm = self.normalizer.transform(X)
        X_norm = np.log1p(X_norm * 1e4)
        X_lsi = self.pcaTransformer.transform(X_norm)
        X_lsi -= X_lsi.mean(axis=1, keepdims=True)
        X_lsi /= X_lsi.std(axis=1, ddof=1, keepdims=True)
        lsi_df = pd.DataFrame(X_lsi, index=adata_use.obs_names).iloc[
            :, int(self.drop_first) :
        ]
        return lsi_df

    def fit_transform(self, adata):
        self.fit(adata)
        return self.transform(adata)


def lsi(
    adata: anndata.AnnData,
    n_components: int = 20,
    use_highly_variable: Optional[bool] = None,
    **kwargs,
) -> None:
    """
    LSI analysis (following the Seurat v3 approach)
    Parameters
    ----------
    adata
        Input dataset
    n_components
        Number of dimensions to use
    use_highly_variable
        Whether to use highly variable features only, stored in
        ``adata.var['highly_variable']``. By default uses them if they
        have been determined beforehand.
    **kwargs
        Additional keyword arguments are passed to
        :func:`sklearn.utils.extmath.randomized_svd`
    """
    if use_highly_variable is None:
        use_highly_variable = "highly_variable" in adata.var
    adata_use = adata[:, adata.var["highly_variable"]] if use_highly_variable else adata
    X = tfidf(adata_use.X)
    X_norm = sklearn.preprocessing.Normalizer(norm="l1").fit_transform(X)
    X_norm = np.log1p(X_norm * 1e4)
    X_lsi = sklearn.utils.extmath.randomized_svd(
        X_norm, n_components, random_state=777, **kwargs
    )[0]
    X_lsi -= X_lsi.mean(axis=1, keepdims=True)
    X_lsi /= X_lsi.std(axis=1, ddof=1, keepdims=True)
    adata.obsm["X_lsi"] = X_lsi


def harmony(df_list, use_gpu=True):
    """
    Harmony batch effect correction applied jointly to multiple dataframes
    """
    all = pd.concat(df_list)
    all_batches = all.pop("batch")
    all_batches.columns = ["batch"]
    all_batches = all_batches.to_frame()
    all_harmony = harmonize(
        all.to_numpy(), all_batches, batch_key="batch", use_gpu=use_gpu, verbose=True
    )
    out_df_list = []
    curr_idx = 0
    for df in df_list:
        out_df_list.append(all_harmony[curr_idx : curr_idx + len(df)])
        curr_idx += len(df)
    return out_df_list



"""Install package."""
import io
import re

from pkg_resources import parse_requirements
from setuptools import find_packages, setup


def read_version(filepath: str) -> str:
    """Read the __version__ variable from the file.

    Args:
        filepath: probably the path to the root __init__.py

    Returns:
        the version
    """
    match = re.search(
        r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        io.open(filepath, encoding="utf_8_sig").read(),
    )
    if match is None:
        raise SystemExit("Version number not found.")
    return match.group(1)


# ease installation during development
vcs = re.compile(r"(git|svn|hg|bzr)\+")
try:
    with open("requirements.txt") as fp:
        VCS_REQUIREMENTS = [
            str(requirement)
            for requirement in parse_requirements(fp)
            if vcs.search(str(requirement))
        ]
except FileNotFoundError:
    # requires verbose flags to show
    print("requirements.txt not found.")
    VCS_REQUIREMENTS = []

setup(
    name="matchclot",
    version=read_version("matchclot/__init__.py"),  # single place for version
    description="Installable matchclot package.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/AI4SCR/MatchCLOT",
    author="Federico Gossi",
    author_email="fgossi@ethz.ch",
    # the following exclusion is to prevent shipping of tests.
    # if you do include them, add pytest to the required packages.
    packages=find_packages(".", exclude=["*tests*"]),
    package_data={"matchclot": ["py.typed"]},
    entry_points="""
        [console_scripts]
        salutation=matchclot.complex_module.core:formal_introduction
    """,
    scripts=[],
    extras_require={
        "vcs": VCS_REQUIREMENTS,
        "test": ["pytest", "pytest-cov"],
        "dev": [
            # tests
            "pytest",
            "pytest-cov",
            # checks
            "black==22.3.0",
            "flake8",
            "mypy",
            # docs
            "sphinx",
            "sphinx-autodoc-typehints",
            "better-apidoc",
            "six",
            "sphinx_rtd_theme",
            "myst-parser",
        ],
    },
    install_requires=[
        # versions should be very loose here, just exclude unsuitable versions
        # because your dependencies also have dependencies and so on ...
        # being too strict here will make dependency resolution harder
        "click",
    ],
)



import argparse
import os
import pickle

import anndata as ad
import numpy as np
import pandas as pd
import scipy.sparse
import torch
from distutils.util import strtobool

from catalyst.utils import set_global_seed

from matchclot.embedding.models import Modality_CLIP, Encoder
from matchclot.matching.matching import OT_matching, MWB_matching
from matchclot.preprocessing.preprocess import harmony
from matchclot.run.evaluate import evaluate
from matchclot.utils.dataloaders import ModalityMatchingDataset
from matchclot.utils.hyperparameters import (
    defaults_common,
    defaults_GEX2ADT,
    defaults_GEX2ATAC,
    baseline_GEX2ATAC,
    baseline_GEX2ADT,
)

# sys.path.append(".")
if __name__ == "__main__":
    os.environ["CUDA_LAUNCH_BLOCKING"] = "1"
    # setting device on GPU if available, else CPU
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print("Using device:", device)

    # Define argument parsers
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="TASK")

    # Common args
    for key, value in defaults_common.items():
        parser.add_argument(
            "--" + key,
            default=value,
            type=(lambda x: bool(strtobool(x))) if type(value) == bool else type(value),
        )

    # GEX2ADT args
    parser_GEX2ADT = subparsers.add_parser("GEX2ADT", help="train GEX2ADT model")
    for key, value in defaults_GEX2ADT.items():
        parser_GEX2ADT.add_argument("--" + key, default=value, type=type(value))

    # GEX2ATAC args
    parser_GEX2ATAC = subparsers.add_parser("GEX2ATAC", help="train GEX2ATAC model")
    for key, value in defaults_GEX2ATAC.items():
        parser_GEX2ATAC.add_argument("--" + key, default=value, type=type(value))

    # Parse args
    args, unknown_args = parser.parse_known_args()

    # Set global random seed
    set_global_seed(args.SEED)

    # Define file paths
    if args.TASK == "GEX2ADT":
        dataset_path = os.path.join(
            args.DATASETS_PATH,
            "openproblems_bmmc_cite_phase2_rna/"
            "openproblems_bmmc_cite_phase2_rna.censor_dataset.output_",
        )
        pretrain_path = os.path.join(args.PRETRAIN_PATH, "GEX2ADT")
        is_multiome = False
    elif args.TASK == "GEX2ATAC":
        dataset_path = os.path.join(
            args.DATASETS_PATH,
            "openproblems_bmmc_multiome_phase2_rna/"
            "openproblems_bmmc_multiome_phase2_rna.censor_dataset.output_",
        )
        pretrain_path = os.path.join(args.PRETRAIN_PATH, "GEX2ATAC")
        is_multiome = True
    else:
        raise ValueError("Unknown task: " + args.TASK)
    if args.CUSTOM_DATASET_PATH != "":
        dataset_path = args.CUSTOM_DATASET_PATH
        assert args.TRANSDUCTIVE is False
        assert args.HARMONY is False

    par = {
        "input_train_mod1": f"{dataset_path}train_mod1.h5ad",
        "input_train_mod2": f"{dataset_path}train_mod2.h5ad",
        "input_test_mod1": f"{dataset_path}test_mod1.h5ad",
        "input_test_mod2": f"{dataset_path}test_mod2.h5ad",
        "input_pretrain": pretrain_path,
        "output": os.path.join(args.PRETRAIN_PATH, args.OUT_NAME + args.TASK),
    }

    # Overwrite configurations for ablation study
    if args.HYPERPARAMS is False:
        if is_multiome:
            for hyperparam, baseline_value in baseline_GEX2ATAC.items():
                setattr(args, hyperparam, baseline_value)
        else:
            for hyperparam, baseline_value in baseline_GEX2ADT.items():
                setattr(args, hyperparam, baseline_value)
    print("args:", args, "unknown_args:", unknown_args)

    # Load data
    if args.TRANSDUCTIVE:
        input_train_mod1 = ad.read_h5ad(par["input_train_mod1"])
        input_train_mod2 = ad.read_h5ad(par["input_train_mod2"])
    input_test_mod1 = ad.read_h5ad(par["input_test_mod1"])
    input_test_mod2 = ad.read_h5ad(par["input_test_mod2"])

    N_cells = input_test_mod1.shape[0]
    print(
        "mod1 cells:", input_test_mod1.shape[0], "mod2 cells:", input_test_mod2.shape[0]
    )
    assert input_test_mod2.shape[0] == N_cells

    # Load and apply LSI transformation
    with open(par["input_pretrain"] + "/lsi_GEX_transformer.pickle", "rb") as f:
        lsi_transformer_gex = pickle.load(f)
    if is_multiome:
        with open(par["input_pretrain"] + "/lsi_ATAC_transformer.pickle", "rb") as f:
            lsi_transformer_atac = pickle.load(f)
        if args.TRANSDUCTIVE:
            gex_train = lsi_transformer_gex.transform(input_train_mod1)
            mod2_train = lsi_transformer_atac.transform(input_train_mod2)
        gex_test = lsi_transformer_gex.transform(input_test_mod1)
        mod2_test = lsi_transformer_atac.transform(input_test_mod2)
    else:
        if args.TRANSDUCTIVE:
            gex_train = lsi_transformer_gex.transform(input_train_mod1)
            mod2_train = input_train_mod2.to_df()
        gex_test = lsi_transformer_gex.transform(input_test_mod1)
        mod2_test = input_test_mod2.to_df()

    if args.HARMONY:
        # Apply Harmony batch effect correction
        gex_test["batch"] = input_test_mod1.obs.batch
        mod2_test["batch"] = input_test_mod2.obs.batch

        if args.TRANSDUCTIVE:
            # Transductive setting
            gex_train["batch"] = input_train_mod1.obs.batch
            (
                gex_train,
                gex_test,
            ) = harmony([gex_train, gex_test])
            mod2_train["batch"] = input_train_mod2.obs.batch
            (
                mod2_train,
                mod2_test,
            ) = harmony([mod2_train, mod2_test])
        else:
            (gex_test,) = harmony([gex_test])
            (mod2_test,) = harmony([mod2_test])

    # Load pretrained models and ensemble predictions
    sim_matrix = np.zeros((N_cells, N_cells))
    for fold in range(0, 9):
        weight_file = par["input_pretrain"] + "/" + str(fold) + "/model.best.pth"
        if os.path.exists(weight_file):
            print("Loading weights from " + weight_file)
            weight = torch.load(weight_file, map_location="cpu")

            # Define modality encoders
            if is_multiome:
                model = Modality_CLIP(
                    Encoder=Encoder,
                    layers_dims=(
                        [args.LAYERS_DIM_ATAC],
                        [args.LAYERS_DIM_GEX0, args.LAYERS_DIM_GEX1],
                    ),
                    dropout_rates=(
                        [args.DROPOUT_RATES_ATAC],
                        [args.DROPOUT_RATES_GEX0, args.DROPOUT_RATES_GEX1],
                    ),
                    dim_mod1=args.N_LSI_COMPONENTS_ATAC,
                    dim_mod2=args.N_LSI_COMPONENTS_GEX,
                    output_dim=args.EMBEDDING_DIM,
                    T=args.LOG_T,
                    noise_amount=args.SFA_NOISE,
                ).to(device)
            else:
                model = Modality_CLIP(
                    Encoder=Encoder,
                    layers_dims=(
                        [args.LAYERS_DIM_ADT0, args.LAYERS_DIM_ADT1],
                        [args.LAYERS_DIM_GEX0, args.LAYERS_DIM_GEX1],
                    ),
                    dropout_rates=(
                        [args.DROPOUT_RATES_ADT0, args.DROPOUT_RATES_ADT1],
                        [args.DROPOUT_RATES_GEX0, args.DROPOUT_RATES_GEX1],
                    ),
                    dim_mod1=args.N_LSI_COMPONENTS_ADT,
                    dim_mod2=args.N_LSI_COMPONENTS_GEX,
                    output_dim=args.EMBEDDING_DIM,
                    T=args.LOG_T,
                    noise_amount=args.SFA_NOISE,
                ).to(device)

            # Load pretrained weights
            model.load_state_dict(weight)

            # Load torch datasets
            dataset_test = ModalityMatchingDataset(
                pd.DataFrame(gex_test), pd.DataFrame(mod2_test)
            )
            data_test = torch.utils.data.DataLoader(dataset_test, 32, shuffle=False)

            # Predict on test set
            all_emb_mod1 = []
            all_emb_mod2 = []
            indexes = []
            model.eval()
            for batch in data_test:
                x1 = batch["features_first"].float()
                x2 = batch["features_second"].float()
                # The model applies the GEX encoder to the second argument, here x1
                logits, features_mod2, features_mod1 = model(
                    x2.to(device), x1.to(device)
                )

                all_emb_mod1.append(features_mod1.detach().cpu())
                all_emb_mod2.append(features_mod2.detach().cpu())

            all_emb_mod1 = torch.cat(all_emb_mod1)
            all_emb_mod2 = torch.cat(all_emb_mod2)

            # Save the embeddings concatenated according to the true order and predicted order
            if args.SAVE_EMBEDDINGS:
                # Assumes that the two modalities have the cells in the same order
                all_emb_mod12_true = torch.cat((all_emb_mod1, all_emb_mod2), dim=1)
                file = par["output"] + "emb_mod12_fold" + str(fold) + "_truematch.pt"
                torch.save(all_emb_mod12_true, file)
                print("Prediction saved to", file)
                predicted_match = ad.read_h5ad(par["output"] + ".h5ad")
                Xsol = torch.tensor(predicted_match.X)
                all_emb_mod12_pred = torch.cat(
                    (all_emb_mod1, all_emb_mod2[Xsol.argmax(1)]), dim=1
                )
                file = par["output"] + "emb_mod12_fold" + str(fold) + "_predmatch.pt"
                torch.save(all_emb_mod12_pred, file)
                print("Prediction saved to", file)

            # Calculate the cosine similarity matrix and add it to the ensemble
            sim_matrix += (all_emb_mod1 @ all_emb_mod2.T).detach().cpu().numpy()

    # save the full similarity matrix
    # np.save("similarity_matrix.npy", sim_matrix)

    if args.BATCH_LABEL_MATCHING:
        # Split matching by batch label
        mod1_splits = set(input_test_mod1.obs["batch"])
        mod2_splits = set(input_test_mod2.obs["batch"])
        splits = mod1_splits | mod2_splits
        matching_matrices, mod1_obs_names, mod2_obs_names = [], [], []
        mod1_obs_index = input_test_mod1.obs.index
        mod2_obs_index = input_test_mod2.obs.index

        for split in splits:
            print("matching split", split)
            mod1_split = input_test_mod1[input_test_mod1.obs["batch"] == split]
            mod2_split = input_test_mod2[input_test_mod2.obs["batch"] == split]
            mod1_obs_names.append(mod1_split.obs_names)
            mod2_obs_names.append(mod2_split.obs_names)
            mod1_indexes = mod1_obs_index.get_indexer(mod1_split.obs_names)
            mod2_indexes = mod2_obs_index.get_indexer(mod2_split.obs_names)
            sim_matrix_split = sim_matrix[np.ix_(mod1_indexes, mod2_indexes)]
            # save the split similarity matrix
            # np.save("test_similarity_matrix"+str(split)+".npy", sim_matrix_split)

            if args.OT_MATCHING:
                # Compute OT matching
                matching_matrices.append(
                    OT_matching(sim_matrix_split, entropy_reg=args.OT_ENTROPY)
                )
            else:
                # Max-weight bipartite matching
                matching_matrices.append(MWB_matching(sim_matrix_split))

        # Assemble the matching matrices and reorder according to the original order of cell profiles
        matching_matrix = scipy.sparse.block_diag(matching_matrices, format="csc")
        mod1_obs_names = pd.Index(np.concatenate(mod1_obs_names))
        mod2_obs_names = pd.Index(np.concatenate(mod2_obs_names))
        matching_matrix = matching_matrix[
            mod1_obs_names.get_indexer(mod1_obs_index), :
        ][:, mod2_obs_names.get_indexer(mod2_obs_index)]
    else:
        if args.OT_MATCHING:
            # Compute OT matching
            matching_matrix = OT_matching(sim_matrix, entropy_reg=args.OT_ENTROPY)
        else:
            # Max-weight bipartite matching
            matching_matrix = MWB_matching(sim_matrix)
        matching_matrix = pd.DataFrame(matching_matrix)

    out = ad.AnnData(
        X=matching_matrix,
        uns={
            # "dataset_id": input_test_mod1.uns["dataset_id"],
            "method_id": "MatchCLOT",
        },
    )

    # Save the matching matrix
    out.write_h5ad(par["output"] + ".h5ad", compression="gzip")
    print("Prediction saved to", par["output"] + ".h5ad")

    # Load the solution for evaluation
    if is_multiome:
        sol_path = os.path.join(
            args.DATASETS_PATH,
            "openproblems_bmmc_multiome_phase2_rna"
            "/openproblems_bmmc_multiome_phase2_rna.censor_dataset.output_",
        )
    else:
        sol_path = os.path.join(
            args.DATASETS_PATH,
            "openproblems_bmmc_cite_phase2_rna/openproblems_bmmc_cite_phase2_rna"
            ".censor_dataset.output_",
        )
    sol = ad.read_h5ad(sol_path + "test_sol.h5ad")

    # Score the prediction and save the results
    scores_path = os.path.join("scores", args.OUT_NAME + args.TASK + ".txt")
    evaluate(out, sol, scores_path)



import torch
import torch.nn.functional as F
from torch import nn


class BatchSwapNoise(nn.Module):
    """Swap Noise module"""

    def __init__(self, p):
        super().__init__()
        self.p = p

    def forward(self, x):
        if self.training:
            mask = torch.rand(x.size()) > (1 - self.p)
            idx = torch.add(
                torch.arange(x.nelement()),
                (
                    torch.floor(torch.rand(x.size()) * x.size(0)).type(torch.LongTensor)
                    * (mask.type(torch.LongTensor) * x.size(1))
                ).view(-1),
            )
            idx[idx >= x.nelement()] = idx[idx >= x.nelement()] - x.nelement()
            return x.view(-1)[idx].view(x.size())
        else:
            return x


class Encoder(nn.Module):
    """Single modality encoder MLP with dropout and stochastic feature augmentation (SFA)
    https://openaccess.thecvf.com/content/ICCV2021/papers
    /Li_A_Simple_Feature_Augmentation_for_Domain_Generalization_ICCV_2021_paper.pdf"""

    def __init__(
        self,
        n_input,
        embedding_size,
        dropout_rates,
        dims_layers,
        swap_noise_ratio,
        noise_amount=0.0,
    ):
        super(Encoder, self).__init__()
        dropout = []
        layers = [nn.Linear(n_input, dims_layers[0])]

        for i in range(len(dims_layers) - 1):
            layers.append(nn.Linear(dims_layers[i], dims_layers[i + 1]))
        for i in range(len(dropout_rates)):
            dropout.append(nn.Dropout(p=dropout_rates[i]))

        layers.append(nn.Linear(dims_layers[-1], embedding_size))

        self.fc_list = nn.ModuleList(layers)
        print("dropout list", dropout)
        self.dropout_list = nn.ModuleList(dropout)
        self.noise_amount = noise_amount
        print("SFA with noise:", noise_amount)

    def forward(self, x):
        for i in range(len(self.fc_list) - 1):
            if i > 0 and self.training and self.noise_amount > 0:
                x = torch.mul(
                    x,
                    torch.ones_like(x)
                    + self.noise_amount * torch.randn_like(x, device=x.device),
                )
                x += self.noise_amount * torch.randn_like(x, device=x.device)
            x = F.elu(self.fc_list[i](x))
            if i < len(self.dropout_list):
                x = self.dropout_list[i](x)

        x = self.fc_list[-1](x)
        return x


class Modality_CLIP(nn.Module):
    """CLIP-inspired architecture"""

    def __init__(
        self,
        Encoder,
        layers_dims,
        dropout_rates,
        dim_mod1,
        dim_mod2,
        output_dim,
        T,
        swap_rate_1=0.0,
        swap_rate_2=0.0,
        noise_amount=0.0,
    ):
        super(Modality_CLIP, self).__init__()

        self.encoder_modality1 = Encoder(
            dim_mod1,
            output_dim,
            dropout_rates[0],
            layers_dims[0],
            swap_rate_1,
            noise_amount=noise_amount,
        )
        self.encoder_modality2 = Encoder(
            dim_mod2,
            output_dim,
            dropout_rates[1],
            layers_dims[1],
            swap_rate_2,
            noise_amount=noise_amount,
        )

        self.logit_scale = nn.Parameter(torch.ones([]) * T)

    def forward(self, features_first, features_second):
        features_mod1 = self.encoder_modality1(features_first)
        features_mod2 = self.encoder_modality2(features_second)

        features_mod1 = features_mod1 / torch.norm(
            features_mod1, p=2, dim=-1, keepdim=True
        )
        features_mod2 = features_mod2 / torch.norm(
            features_mod2, p=2, dim=-1, keepdim=True
        )

        logit_scale = self.logit_scale.exp()

        logits = logit_scale * features_mod1 @ features_mod2.T

        return logits, features_mod1, features_mod2


def symmetric_npair_loss(logits, targets):
    """CLIP loss"""
    loss = 0.5 * (F.cross_entropy(logits, targets) + F.cross_entropy(logits.T, targets))
    return loss



import torch
from catalyst import runners, metrics, dl

from matchclot.embedding.models import symmetric_npair_loss


class scRNARunner(runners.Runner):
    def handle_batch(self, batch):
        # model train/valid step
        # unpack the batch
        features_first = batch["features_first"].float()
        features_second = batch["features_second"].float()

        # run model forward pass
        logits, embeddings_first, embeddings_second = self.model(
            features_first, features_second
        )
        targets = torch.arange(logits.shape[0]).to(logits.device)

        # compute the loss
        loss = symmetric_npair_loss(logits, targets)

        # log metrics
        batch_temperature = self.model.logit_scale.exp().item()
        self.batch_metrics.update({"loss": loss})
        self.batch_metrics.update({"T": batch_temperature})

        self.batch = {
            "features_first": features_first,
            "features_second": features_second,
            "embeddings_first": embeddings_first,
            "embeddings_second": embeddings_second,
            "scores": logits,
            "targets": targets,
            "temperature": batch_temperature,
        }
        self.input = {
            "features_first": features_first,
            "features_second": features_second,
        }
        self.output = {
            "scores": logits,
            "embeddings_first": embeddings_first,
            "embeddings_second": embeddings_second,
        }

        # run model backward pass
        if self.is_train_loader:
            self.engine.backward(loss)
            self.optimizer.step()
            self.optimizer.zero_grad()

    def get_loggers(self):
        return {
            "console": dl.ConsoleLogger(),
        }


class CustomMetric(metrics.ICallbackLoaderMetric):
    """Top1, Top5 accuracy metrics and competition score, without applying the matching algorithm."""

    def __init__(
        self, compute_on_call: bool = True, prefix: str = None, suffix: str = None
    ):
        super().__init__(compute_on_call=compute_on_call)
        self.prefix = prefix or ""
        self.suffix = suffix or ""
        self.embeddings_list_first = []
        self.embeddings_list_second = []
        self.batch_size = 256  # For batched computation of metrics
        self.extended_statistics = False

    def reset(self, num_batches: int, num_samples: int) -> None:
        self.embeddings_list_first = []
        self.embeddings_list_second = []
        torch.cuda.empty_cache()

    def update(self, *args, **kwargs) -> None:
        embeddings_first = kwargs["embeddings_first"]
        embeddings_second = kwargs["embeddings_second"]
        temperature = kwargs["temperature"]
        self.embeddings_list_first.append(temperature * embeddings_first)
        self.embeddings_list_second.append(embeddings_second)

    def compute(self):
        raise NotImplementedError("This method is not supported")

    def compute_key_value(self):
        all_embeddings_first = torch.cat(self.embeddings_list_first).detach().cpu()
        all_embeddings_second = torch.cat(self.embeddings_list_second).detach().cpu()
        N = all_embeddings_first.shape[0]

        # print("Calculating metrics")
        embeddings_first = all_embeddings_first
        embeddings_second = all_embeddings_second
        logits = embeddings_first @ embeddings_second.T
        del embeddings_first
        del embeddings_second
        labels = torch.arange(logits.shape[0])

        forward_accuracy = 0
        for i in range(0, N, self.batch_size):
            curr_batch_size = min(self.batch_size, N - i)
            logits_batch = logits[i : i + curr_batch_size, :]  # row batch
            forward_accuracy += (
                torch.argmax(logits_batch, dim=1) + i == labels[i : i + curr_batch_size]
            ).float().mean().item() / curr_batch_size
            del logits_batch
        backward_accuracy = 0
        for i in range(0, N, self.batch_size):
            curr_batch_size = min(self.batch_size, N - i)
            logits_batch = logits[:, i : i + curr_batch_size]  # column batch
            backward_accuracy += (
                torch.argmax(logits_batch, dim=0) + i == labels[i : i + curr_batch_size]
            ).float().mean().item() / curr_batch_size
            del logits_batch
        avg_accuracy = 0.5 * (forward_accuracy + backward_accuracy)

        if self.extended_statistics:
            top1_competition_metric = 0
            for i in range(0, N, self.batch_size):
                curr_batch_size = min(self.batch_size, N - i)
                logits_batch = logits[i : i + curr_batch_size, :]  # row batch
                logits_row_sums = logits_batch.clip(min=0).sum(dim=1)
                top1_competition_metric += (
                    logits_batch.clip(min=0)
                    .diagonal(offset=i)
                    .div(logits_row_sums)
                    .mean()
                    .item()
                    / curr_batch_size
                )

            _, top_indexes_forward = logits.topk(5, dim=1)
            _, top_indexes_backward = logits.topk(5, dim=0)
            l_forward = labels.expand(5, logits.shape[0]).T
            del logits
            l_backward = l_forward.T
            top5_forward_accuracy = (
                torch.any(top_indexes_forward == l_forward, 1).float().mean().item()
            )
            top5_backward_accuracy = (
                torch.any(top_indexes_backward == l_backward, 0).float().mean().item()
            )
            top5_avg_accuracy = 0.5 * (top5_forward_accuracy + top5_backward_accuracy)

        loader_metrics = {
            "forward_acc": forward_accuracy,
            "backward_acc": backward_accuracy,
            "avg_acc": avg_accuracy,
        }
        if self.extended_statistics:
            loader_metrics.update(
                {
                    "top1_competition_metric": top1_competition_metric,
                    "top5_forward_acc": top5_forward_accuracy,
                    "top5_backward_acc": top5_backward_accuracy,
                    "top5_avg_acc": top5_avg_accuracy,
                }
            )
        return loader_metrics
